==> ver.B.b11.02.tar.gz: ver.B.b11.02/env/src/memory_debug/memory_debug.c <==
/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#define MEMORY_DEBUG
/*
#define DONT_USE_LOCK
*/
#include	<stdlib.h>
#include	"log.h"
#include	"memory_debug.h"
#include	"lock_level.h"
#include	"save_global.h"
#include	"avt.h"
#include	"utils.h"

#ifndef DONT_USE_LOCK
#include	"task.h"
#include	"pri_level.h"
#endif

#define LOG_MEM_LIST_LEN	20

void memory_debug_task(); 


#ifdef MEMORY_DEBUG


int flag;

#ifndef DONT_USE_LOCK
SEM md_lock;
int md_flag;
#endif /* DONT_USE_LOCK */

#ifdef DONT_USE_LOCK
#define LOCK	{}
#define UNLOCK	{}
#define NEW_LOCK {}
#endif /* DON_USE_LOCK */

#ifndef DONT_USE_LOCK
#define LOCK	{md_flag=1;if ( lock_up_test(md_lock) ) lock_task(md_lock);}
#define UNLOCK	{md_flag=1;if ( lock_up_test(md_lock) ) unlock_task(md_lock,"unlock");}
#define NEW_LOCK	md_lock = new_lock(LL_MD);
//#define LOCK er_panic("LOCK");

#endif /* DONT_USE_LOCK */


typedef struct mem_header {
	MEM_LIST *		mlp;
	int			size;

	char *			last_free_file;
	int			last_free_line;

} MEM_HEADER;


int lm_trace(AVT_NODE * a,AVT_NODE ** ret);
int lm_log_output(AVT_NODE * a,int * cnt);

#define MEM_LIST_HASH_SIZE	53
#define MEM_FREE_INFO_MAX	1000
MEM_LIST ** mem_list_hash;
MEM_LIST * mem_max_list;

MEM_FREE_INFO ** mem_free_info_hash;
MEM_FREE_INFO * mem_free_time_head;
MEM_FREE_INFO * mem_free_time_tail;
int mem_free_info_len;

MEM_PURGE_TBL *	mpt_root;
INTEGER64 total_mem_size;

void ind_tick();
void list_tick();

void
init_d_alloc()
{
	NEW_LOCK
#ifndef DONT_USE_LOCK
	if ( md_flag == 1 )
		er_panic("init_d_alloc");
	new_tick(list_tick,30,0);
	create_task(memory_debug_task,PRI_FETCH,0);
#endif
}

void
_new_mem_hash()
{
int i;
	mem_list_hash = malloc(MEM_LIST_HASH_SIZE*sizeof(MEM_LIST*));
	mem_free_info_hash = malloc(MEM_LIST_HASH_SIZE
		*sizeof(MEM_FREE_INFO*));
	for ( i = 0 ; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		mem_list_hash[i] = 0;
		mem_free_info_hash[i] = 0;
	}
}

MEM_LIST *
_search_mem_list(char * file,int line)
{
MEM_LIST * ret;
int key;
	key = line % MEM_LIST_HASH_SIZE;
	for ( ret = mem_list_hash[key] ; ret ; ret = ret->next )
		if ( ret->line == line &&
			strcmp(ret->file,file) == 0 )
			return ret;
	return 0;
}

MEM_LIST *
_new_mem_list(char * file,int line)
{
MEM_LIST * ret;
int key;
	ret = malloc(sizeof(*ret));
	total_mem_size += sizeof(*ret);
	key = line % MEM_LIST_HASH_SIZE;
	ret->line = line;
	ret->file = file;
	ret->size = 0;
	ret->next = mem_list_hash[key];
	mem_list_hash[key] = ret;
	return ret;
}




void *
xx_d_alloc(int size,char * file,int line)
{
MEM_HEADER * mh;
MEM_LIST * ml;
void * ret;
	LOCK;
	if ( mem_list_hash == 0 )
		_new_mem_hash();
	mh = malloc(size + sizeof(*mh));
if ( mh == 0 )
er_panic("alloc");
	total_mem_size += size + sizeof(*mh);
	if ( mh == 0 ) {
		ret = 0;
	}
	else {
		ml = _search_mem_list(file,line);
		if ( ml == 0 )
			ml = _new_mem_list(file,line);
		mh->mlp = ml;
		mh->size = size;

		mh->last_free_file = 0;
		mh->last_free_line = 0;

		ml->size += size;
		ret = mh+1;
	}
	UNLOCK;
	return ret;
}



void * xx_d_calloc(int size,int el,char * file,int line)
{
char * ret;
int i;
	ret = xx_d_alloc(size*el,file,line);
	if ( ret == 0 )
		return 0;
	for ( i = 0 ; i < size*el ; i ++ )
		ret[i] = 0;
	return ret;
}

void * d_re_alloc(void * ptr,int size)
{
void * ret;
MEM_HEADER * mh;
MEM_LIST *ml;
int old_size;
	LOCK;
	if ( ptr == 0 ) {
		UNLOCK;
		return d_alloc(size); 
	}
	mh = ptr;
	mh = mh - 1;
	ml = mh->mlp;
	old_size = mh->size;
	if ( ml->size < old_size )
		er_panic_x("d_re_alloc");
	mh = realloc(mh,size + sizeof(*mh));
	if ( mh ) {
		ml->size = ml->size - old_size + size;
		mh->size = size;
		ret = mh + 1;
	}
	else {
		ml->size -= old_size;
		ret = 0;
	}
	UNLOCK;
	return ret;
}

void
_delete_free_info(MEM_FREE_INFO * fi)
{
unsigned int key;
MEM_FREE_INFO ** fip;
	key = fi->ptr % MEM_LIST_HASH_SIZE;
	fip = &mem_free_info_hash[key];
	for ( ; *fip && *fip != fi ; fip = &(*fip)->next );
	if ( *fip == 0 )
		er_panic_x("free");
	*fip = fi->next;
	free(fi);
	total_mem_size -= sizeof(*fi);
}

void
_insert_free_info(void * ptr,char * file,int line)
{
unsigned int _ptr;
MEM_FREE_INFO * fi;
unsigned int key;
	_ptr = (unsigned int)ptr;
	fi = malloc(sizeof(*fi));
	total_mem_size += sizeof(*fi);
	fi->ptr = _ptr;
	fi->file = file;
	fi->line = line;

	key = _ptr % MEM_LIST_HASH_SIZE;
	fi->next = mem_free_info_hash[key];
	mem_free_info_hash[key] = fi;

	fi->time_next = 0;
	if ( mem_free_time_tail ) {
		mem_free_time_tail->time_next = fi;
		mem_free_time_tail = fi;
	}
	else mem_free_time_tail = mem_free_time_head = fi;
	mem_free_info_len ++;
	for ( ; mem_free_info_len > MEM_FREE_INFO_MAX ; ) {
		fi = mem_free_time_head;
		mem_free_time_head = fi->time_next;
		_delete_free_info(fi);
		mem_free_info_len --;
	}
}

void xx_d_f_ree(void * ptr,char * file,int line)
{
MEM_HEADER * mh;
MEM_LIST * ml;
int s;
	LOCK;

	mh = ptr;
	mh = mh - 1;
	if ( mh->mlp == 0 || mh->size < 0 ) {
		er_panic_x("xx_d_f_ree(1)");
	}

	if ( mh->last_free_file || mh->last_free_line ) {
		printf("FILE %s %i\n",mh->last_free_file,
			mh->last_free_line);
		er_panic_x("xx_d_f_ree(1)");
	}

//	_insert_free_info(ptr,file,line);

	mh->last_free_file = file;
	mh->last_free_line = line;

	ml = mh->mlp;
	s = mh->size;
	ml->size -= mh->size;
//printf("d_f_ree %p\n",mh);
	{
/*
		int e=GetLastError();
		int en=errno;
*/
		free(mh);
	}

//printf("d_f_ree %p END\n",mh);
	total_mem_size -= s + sizeof(*mh);
	if ( ml->size < 0 )
		er_panic_x("xx_d_f_ree(2)");

	UNLOCK;
}

void
check_mem_header(void * ptr)
{
MEM_HEADER * mh;
MEM_LIST * ml;
	LOCK;

	mh = ptr;
	mh = mh - 1;
/*
	if ( mh->last_free_file || mh->last_free_line ) {
		printf("FILE %s %i\n",mh->last_free_file,
			mh->last_free_line);
		er_panic_x("check_mem_header(1)");
	}
*/
	if ( mh->size < 0 )
		er_panic_x("check_mem_header(3)");
	ml = mh->mlp;
	if ( ml->size < 0 )
		er_panic_x("check_mem_header(4)");
	if ( ml->size - mh->size < 0 )
		er_panic_x("check_mem_header(2)");

	UNLOCK;
}

void
ind_tick()
{
MEM_LIST * ml;
int large;
int i;
	LOCK;
	large = 0;
	for ( i = 0; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		ml = mem_list_hash[i];
		for ( ; ml ; ml = ml->next ) {
			if ( ml->size == 0 )
				continue;
			if ( large < ml->size ) {
				printf("*");
				large = ml->size;
			}
			printf("[%s:%i]==>%i\n",ml->file,ml->line,ml->size);
		}
	}
	printf("IND_MEM END\n");
	UNLOCK;
}


typedef struct log_mem {
	int		size;
	AVT_NODE *	root;
	AVT_NODE	n[LOG_MEM_LIST_LEN];
} LOG_MEM;

void insert_lm(LOG_MEM * lm,MEM_LIST * memlist);

int
lm_cmp(MEM_LIST * m1,MEM_LIST * m2)
{
	if ( m1->size < m2->size )
		return -1;
	if ( m1->size > m2->size )
		return 1;
	if ( ((long)m1) < ((long)m2) )
		return -1;
	if ( ((long)m1) > ((long)m2) )
		return 1;
	return 0;
}

int
lm_trace(AVT_NODE * a,AVT_NODE ** ret)
{
	*ret = a;
	return 1;
}

void
insert_lm(LOG_MEM * lm,MEM_LIST * memlist)
{
AVT_NODE * n;
	if ( lm->size >= LOG_MEM_LIST_LEN ) {
		avt_trace_from_small(lm->root,lm_trace,&n);
		n = avt_delete(&lm->root,n->data,lm_cmp);
	}
	else {
		n = &lm->n[lm->size++];
	}
	n->data = memlist;
	avt_insert(&lm->root,n,lm_cmp);
}

int
lm_log_output(AVT_NODE * a,int * cnt)
{
MEM_LIST * ml;
	ml = a->data;
	log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"::MEM_%i %s:%i / size=%i\n",
		(*cnt)++,
		ml->file,
		ml->line,
		ml->size);
	return 0;
}


void
list_tick()
{
MEM_LIST * ml;
int large;
int i;
char * buf;
int all_size;

LOG_MEM * lm;
//extern int obj_mem_size;

	lm = d_alloc(sizeof(*lm));
	lm->size = 0;
	lm->root = 0;
	LOCK;
	large = 0;
	mem_max_list = 0;
	all_size = 0;
	for ( i = 0; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		ml = mem_list_hash[i];
		for ( ; ml ; ml = ml->next ) {
			if ( ml->size == 0 )
				continue;
			insert_lm(lm,ml);
			all_size += ml->size;
			if ( large < ml->size ) {
				large = ml->size;
				ml->max_next = mem_max_list;
				mem_max_list = ml;
			}
		}
	}

	if ( mem_max_list ) {
		buf = malloc(1000);
		sprintf(buf,"MAX MEMORY = %s:%i / size= %i [TOTAL %i]\n",
			mem_max_list->file,
			mem_max_list->line,
			mem_max_list->size,
			all_size);
	}
	else buf = 0;

	buf = 0;

	UNLOCK;

	if ( buf ) {
		i = 0;
		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"%s",buf);
		free(buf);
		avt_trace_from_large(lm->root,lm_log_output,&i);

		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,
				"IMPREMENTED MEMORY SIZE = %i\n",
					get_mem_size());
//		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,
//				"OBJECT MEMEORY SIZE = %i\n",obj_mem_size);

		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"MEM_END\n");
	}
	
	d_f_ree(lm);
}

MEM_LIST *
get_mem_debug_list(MEM_LIST * m)
{
int key;
MEM_LIST * ret;
	LOCK;
	if ( mem_list_hash == 0 ) {
		goto end;
	}
	if ( m == 0 )
		key = 0;
	else {
		if ( m->next ) {
			ret = m->next;
			goto end;
		}
		key = (m->line % MEM_LIST_HASH_SIZE)+1;
	}
	if ( key >= MEM_LIST_HASH_SIZE ) {
		ret = 0;
		goto end;
	}
	for ( ; key < MEM_LIST_HASH_SIZE &&
			mem_list_hash[key] == 0 ; key ++ );
	if ( key >= MEM_LIST_HASH_SIZE ) {
		ret = 0;
		goto end;
	}
	ret = mem_list_hash[key];
end:
	UNLOCK;
	return ret;
}

#endif /* MEMORY_DEBUG */

#ifndef MEMORY_DEBUG


void
init_d_alloc()
{
}

void *
xx_d_alloc(int size,int d,char * file,int line)
{
	return malloc(size);
}

void *
xx_d_calloc(int size,int el,int d,char * file,int line)
{
	return calloc(size,el);
}

void *
d_re_alloc(void * ptr,int size)
{
	return realloc(ptr,size);
}

void
xx_d_f_ree(void * ptr,char * file,int line)
{
	free(ptr);
}

MEM_LIST * 
get_mem_debug_list(MEM_LIST * m)
{
	return 0;
}


#endif /* ndef MEMORY_DEBUG */



void
set_mem_purge(MEM_PURGE_TBL * t)
{
	t->next = mpt_root;
	mpt_root = t;
}


INTEGER64
get_d_used_memory()
{
	return total_mem_size;
}

#ifndef DONT_USE_LOCK
void
memory_debug_task()
{
INTEGER64 max,min;
INTEGER64 m;
INTEGER64 sz;
MEM_PURGE_TBL * tbl,* target;
int last_tim;
int tim;
	for ( ; ; ) {
		m = get_mem_size();
		sz = get_d_used_memory();

/*
			log_printf(LOG_DEBUG,LOG_LAYER_GB,0,
				"IMPREMENTED MEMORY SIZE = " I64_FORMAT 
						"\n",m);
			log_printf(LOG_DEBUG,LOG_LAYER_GB,0,
				"OBJECT MEMEORY SIZE = " I64_FORMAT "\n",sz);
*/
/*
		if ( m > 256*1024*1024 )
			m = 256*1024*1024;
*/

		max = (m - 25*1024*1024)/2;
		min = (m - 25*1024*1024)/3;

		sleep_sec(5);
//printf("MDT %lli %lli %lli %lli\n",m,max,min,sz);
		if ( sz < max )
			continue;
		for ( ; sz > min ; ) {
			tbl = mpt_root;
			target = 0;
			last_tim = 0;
			for ( ; tbl ; tbl = tbl->next )
				if ( (*tbl->last_access)(&tim) == 0 ) {
					if ( target == 0 || last_tim > tim ) {
						last_tim = tim;
						target = tbl;
					}
				}
			if ( target == 0 )
				break;
			(*target->purge_mem)((sz-min)/3);
		}
	}
}
#endif

==> ver.B.b16.01.tar.gz: ver.B.b16.01/env/src/memory_debug/memory_debug.c <==
/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#define MEMORY_DEBUG
/*
#define DONT_USE_LOCK
*/
#include	<stdlib.h>
#include	"log.h"
#include	"memory_debug.h"
#include	"lock_level.h"
#include	"save_global.h"
#include	"avt.h"
#include	"utils.h"

#ifndef DONT_USE_LOCK
#include	"task.h"
#include	"pri_level.h"
#endif

#define LOG_MEM_LIST_LEN	20

void memory_debug_task(); 


#ifdef MEMORY_DEBUG


int flag;

#ifndef DONT_USE_LOCK
SEM md_lock;
int md_flag;
#endif /* DONT_USE_LOCK */

#ifdef DONT_USE_LOCK
#define LOCK	{}
#define UNLOCK	{}
#define NEW_LOCK {}
#endif /* DON_USE_LOCK */

#ifndef DONT_USE_LOCK
#define LOCK	{md_flag=1;if ( lock_up_test(md_lock) ) lock_task(md_lock);}
#define UNLOCK	{md_flag=1;if ( lock_up_test(md_lock) ) unlock_task(md_lock,"unlock");}
#define NEW_LOCK	md_lock = new_lock(LL_MD);
//#define LOCK er_panic("LOCK");

#endif /* DONT_USE_LOCK */


typedef struct mem_header {
	MEM_LIST *		mlp;
	int			size;

	char *			last_free_file;
	int			last_free_line;

} MEM_HEADER;


int lm_trace(AVT_NODE * a,AVT_NODE ** ret);
int lm_log_output(AVT_NODE * a,int * cnt);

#define MEM_LIST_HASH_SIZE	53
#define MEM_FREE_INFO_MAX	1000
MEM_LIST ** mem_list_hash;
MEM_LIST * mem_max_list;

MEM_FREE_INFO ** mem_free_info_hash;
MEM_FREE_INFO * mem_free_time_head;
MEM_FREE_INFO * mem_free_time_tail;
int mem_free_info_len;

MEM_PURGE_TBL *	mpt_root;
static INTEGER64 total_mem_size;

void ind_tick();
void list_tick();

void
init_d_alloc()
{
	NEW_LOCK
#ifndef DONT_USE_LOCK
	if ( md_flag == 1 )
		er_panic("init_d_alloc");
	new_tick(list_tick,30,0);
	create_task(memory_debug_task,PRI_FETCH,0);
#endif
}

void
_new_mem_hash()
{
int i;
	mem_list_hash = malloc(MEM_LIST_HASH_SIZE*sizeof(MEM_LIST*));
	mem_free_info_hash = malloc(MEM_LIST_HASH_SIZE
		*sizeof(MEM_FREE_INFO*));
	for ( i = 0 ; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		mem_list_hash[i] = 0;
		mem_free_info_hash[i] = 0;
	}
}

MEM_LIST *
_search_mem_list(char * file,int line)
{
MEM_LIST * ret;
int key;
	key = line % MEM_LIST_HASH_SIZE;
	for ( ret = mem_list_hash[key] ; ret ; ret = ret->next )
		if ( ret->line == line &&
			strcmp(ret->file,file) == 0 )
			return ret;
	return 0;
}

MEM_LIST *
_new_mem_list(char * file,int line)
{
MEM_LIST * ret;
int key;
	ret = malloc(sizeof(*ret));
	total_mem_size += sizeof(*ret);
	key = line % MEM_LIST_HASH_SIZE;
	ret->line = line;
	ret->file = file;
	ret->size = 0;
	ret->next = mem_list_hash[key];
	mem_list_hash[key] = ret;
	return ret;
}




void *
xx_d_alloc(int size,char * file,int line)
{
MEM_HEADER * mh;
MEM_LIST * ml;
void * ret;
	LOCK;
	if ( mem_list_hash == 0 )
		_new_mem_hash();
	mh = malloc(size + sizeof(*mh));
	if ( mh == 0 ) {
	  er_panic("alloc");
	}
	total_mem_size += size + sizeof(*mh);
	if ( mh == 0 ) {
		ret = 0;
	}
	else {
		ml = _search_mem_list(file,line);
		if ( ml == 0 )
			ml = _new_mem_list(file,line);
		mh->mlp = ml;
		mh->size = size;

		mh->last_free_file = 0;
		mh->last_free_line = 0;

		ml->size += size;
		ret = mh+1;
	}
	UNLOCK;
	return ret;
}



void * xx_d_calloc(int size,int el,char * file,int line)
{
char * ret;
int i;
	ret = xx_d_alloc(size*el,file,line);
	if ( ret == 0 )
		return 0;
	for ( i = 0 ; i < size*el ; i ++ )
		ret[i] = 0;
	return ret;
}

void * d_re_alloc(void * ptr,int size)
{
void * ret;
MEM_HEADER * mh;
MEM_LIST *ml;
int old_size;
	LOCK;
	if ( ptr == 0 ) {
		UNLOCK;
		return d_alloc(size); 
	}
	mh = ptr;
	mh = mh - 1;
	ml = mh->mlp;
	old_size = mh->size;
	if ( ml->size < old_size )
		er_panic_x("d_re_alloc");
	mh = realloc(mh,size + sizeof(*mh));
	if ( mh ) {
		ml->size = ml->size - old_size + size;
		mh->size = size;
		ret = mh + 1;
	}
	else {
		ml->size -= old_size;
		ret = 0;
	}
	total_mem_size += size - old_size;
	UNLOCK;
	return ret;
}

void
_delete_free_info(MEM_FREE_INFO * fi)
{
unsigned int key;
MEM_FREE_INFO ** fip;
	key = fi->ptr % MEM_LIST_HASH_SIZE;
	fip = &mem_free_info_hash[key];
	for ( ; *fip && *fip != fi ; fip = &(*fip)->next );
	if ( *fip == 0 )
		er_panic_x("free");
	*fip = fi->next;
	free(fi);
	total_mem_size -= sizeof(*fi);
}

void
_insert_free_info(void * ptr,char * file,int line)
{
unsigned int _ptr;
MEM_FREE_INFO * fi;
unsigned int key;
	_ptr = (unsigned int)ptr;
	fi = malloc(sizeof(*fi));
	total_mem_size += sizeof(*fi);
	fi->ptr = _ptr;
	fi->file = file;
	fi->line = line;

	key = _ptr % MEM_LIST_HASH_SIZE;
	fi->next = mem_free_info_hash[key];
	mem_free_info_hash[key] = fi;

	fi->time_next = 0;
	if ( mem_free_time_tail ) {
		mem_free_time_tail->time_next = fi;
		mem_free_time_tail = fi;
	}
	else mem_free_time_tail = mem_free_time_head = fi;
	mem_free_info_len ++;
	for ( ; mem_free_info_len > MEM_FREE_INFO_MAX ; ) {
		fi = mem_free_time_head;
		mem_free_time_head = fi->time_next;
		_delete_free_info(fi);
		mem_free_info_len --;
	}
}


void xx_d_f_ree(void * ptr,char * file,int line)
{
MEM_HEADER * mh;
MEM_LIST * ml;
int s;
	LOCK;

	mh = ptr;
	mh = mh - 1;
	if ( mh->mlp == 0 || mh->size < 0 ) {
		er_panic_x("xx_d_f_ree(1)");
	}

	if ( mh->last_free_file || mh->last_free_line ) {
		printf("FILE %s %i\n",mh->last_free_file,
			mh->last_free_line);
		er_panic_x("xx_d_f_ree(1)");
	}

//	_insert_free_info(ptr,file,line);

	mh->last_free_file = file;
	mh->last_free_line = line;

	ml = mh->mlp;
	s = mh->size;
	ml->size -= mh->size;
//printf("d_f_ree %p\n",mh);
	{
/*
		int e=GetLastError();
		int en=errno;
*/
		free(mh);
	}

//printf("d_f_ree %p END\n",mh);
	total_mem_size -= s + sizeof(*mh);
	if ( ml->size < 0 )
		er_panic_x("xx_d_f_ree(2)");


	UNLOCK;
}

void
check_mem_header(void * ptr)
{
MEM_HEADER * mh;
MEM_LIST * ml;
	LOCK;

	mh = ptr;
	mh = mh - 1;
/*
	if ( mh->last_free_file || mh->last_free_line ) {
		printf("FILE %s %i\n",mh->last_free_file,
			mh->last_free_line);
		er_panic_x("check_mem_header(1)");
	}
*/
	if ( mh->size < 0 )
		er_panic_x("check_mem_header(3)");
	ml = mh->mlp;
	if ( ml->size < 0 )
		er_panic_x("check_mem_header(4)");
	if ( ml->size - mh->size < 0 )
		er_panic_x("check_mem_header(2)");

	UNLOCK;
}

void
ind_tick()
{
MEM_LIST * ml;
int large;
int i;
	LOCK;
	large = 0;
	for ( i = 0; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		ml = mem_list_hash[i];
		for ( ; ml ; ml = ml->next ) {
			if ( ml->size == 0 )
				continue;
			if ( large < ml->size ) {
				printf("*");
				large = ml->size;
			}
			printf("[%s:%i]==>%i\n",ml->file,ml->line,ml->size);
		}
	}
	printf("IND_MEM END\n");
	UNLOCK;
}


typedef struct log_mem {
	int		size;
	AVT_NODE *	root;
	AVT_NODE	n[LOG_MEM_LIST_LEN];
} LOG_MEM;

void insert_lm(LOG_MEM * lm,MEM_LIST * memlist);

int
lm_cmp(MEM_LIST * m1,MEM_LIST * m2)
{
	if ( m1->size < m2->size )
		return -1;
	if ( m1->size > m2->size )
		return 1;
	if ( ((long)m1) < ((long)m2) )
		return -1;
	if ( ((long)m1) > ((long)m2) )
		return 1;
	return 0;
}

int
lm_trace(AVT_NODE * a,AVT_NODE ** ret)
{
	*ret = a;
	return 1;
}

void
insert_lm(LOG_MEM * lm,MEM_LIST * memlist)
{
AVT_NODE * n;
	if ( lm->size >= LOG_MEM_LIST_LEN ) {
		avt_trace_from_small(lm->root,lm_trace,&n);
		n = avt_delete(&lm->root,n->data,lm_cmp);
	}
	else {
		n = &lm->n[lm->size++];
	}
	n->data = memlist;
	avt_insert(&lm->root,n,lm_cmp);
}

int
lm_log_output(AVT_NODE * a,int * cnt)
{
MEM_LIST * ml;
	ml = a->data;
	log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"::MEM_%i %s:%i / size=%i\n",
		(*cnt)++,
		ml->file,
		ml->line,
		ml->size);
	return 0;
}


void
list_tick()
{
MEM_LIST * ml;
int large;
int i;
char * buf;
int all_size;

LOG_MEM * lm;
//extern int obj_mem_size;

	lm = d_alloc(sizeof(*lm));
	lm->size = 0;
	lm->root = 0;
	LOCK;
	large = 0;
	mem_max_list = 0;
	all_size = 0;
	for ( i = 0; i < MEM_LIST_HASH_SIZE ; i ++ ) {
		ml = mem_list_hash[i];
		for ( ; ml ; ml = ml->next ) {
			if ( ml->size == 0 )
				continue;
			insert_lm(lm,ml);
			all_size += ml->size;
			if ( large < ml->size ) {
				large = ml->size;
				ml->max_next = mem_max_list;
				mem_max_list = ml;
			}
		}
	}
/*
	if ( mem_max_list ) {

		buf = malloc(1000);
		sprintf(buf,"MAX MEMORY = %s:%i / size= %i [TOTAL %i]\n",
			mem_max_list->file,
			mem_max_list->line,
			mem_max_list->size,
			all_size);

	}
	else buf = 0;
*/
	buf = 0;

	UNLOCK;

	if ( buf ) {
		i = 0;
		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"%s",buf);
		free(buf);
		avt_trace_from_large(lm->root,lm_log_output,&i);

		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,
				"IMPREMENTED MEMORY SIZE = %i\n",
					get_mem_size());
//		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,
//				"OBJECT MEMEORY SIZE = %i\n",obj_mem_size);

		log_printf(LOG_DEBUG,LOG_LAYER_LIB,0,"MEM_END\n");
	}
	
	d_f_ree(lm);
}

MEM_LIST *
get_mem_debug_list(MEM_LIST * m)
{
int key;
MEM_LIST * ret;
	LOCK;
	if ( mem_list_hash == 0 ) {
		goto end;
	}
	if ( m == 0 )
		key = 0;
	else {
		if ( m->next ) {
			ret = m->next;
			goto end;
		}
		key = (m->line % MEM_LIST_HASH_SIZE)+1;
	}
	if ( key >= MEM_LIST_HASH_SIZE ) {
		ret = 0;
		goto end;
	}
	for ( ; key < MEM_LIST_HASH_SIZE &&
			mem_list_hash[key] == 0 ; key ++ );
	if ( key >= MEM_LIST_HASH_SIZE ) {
		ret = 0;
		goto end;
	}
	ret = mem_list_hash[key];
end:
	UNLOCK;
	return ret;
}

#endif /* MEMORY_DEBUG */

#ifndef MEMORY_DEBUG


void
init_d_alloc()
{
}

void *
xx_d_alloc(int size,int d,char * file,int line)
{
	return malloc(size);
}

void *
xx_d_calloc(int size,int el,int d,char * file,int line)
{
	return calloc(size,el);
}

void *
d_re_alloc(void * ptr,int size)
{
	return realloc(ptr,size);
}

void
xx_d_f_ree(void * ptr,char * file,int line)
{
	free(ptr);
}

MEM_LIST * 
get_mem_debug_list(MEM_LIST * m)
{
	return 0;
}


#endif /* ndef MEMORY_DEBUG */



void
set_mem_purge(MEM_PURGE_TBL * t)
{
	t->next = mpt_root;
	mpt_root = t;
}


INTEGER64
get_d_used_memory()
{
	return total_mem_size;
}

#ifndef DONT_USE_LOCK
void
memory_debug_task()
{
INTEGER64 max,min;
INTEGER64 m;
INTEGER64 sz;
MEM_PURGE_TBL * tbl,* target;
int last_tim;
int tim;
	for ( ; ; ) {
		m = get_mem_size();
		sz = get_d_used_memory();

/*
			log_printf(LOG_DEBUG,LOG_LAYER_GB,0,
				"IMPREMENTED MEMORY SIZE = " I64_FORMAT 
						"\n",m);
			log_printf(LOG_DEBUG,LOG_LAYER_GB,0,
				"OBJECT MEMEORY SIZE = " I64_FORMAT "\n",sz);
*/
/*
		if ( m > 256*1024*1024 )
			m = 256*1024*1024;
*/

		max = (m - 25*1024*1024)/2;
		min = (m - 25*1024*1024)/3;

		sleep_sec(5);
//printf("MDT %lli %lli %lli %lli\n",m,max,min,sz);
		if ( sz < max )
			continue;
		for ( ; sz > min ; ) {
			tbl = mpt_root;
			target = 0;
			last_tim = 0;
			for ( ; tbl ; tbl = tbl->next )
				if ( (*tbl->last_access)(&tim) == 0 ) {
					if ( target == 0 || last_tim > tim ) {
						last_tim = tim;
						target = tbl;
					}
				}
			if ( target == 0 )
				break;
			(*target->purge_mem)((sz-min)/3);
		}
	}
}
#endif
